@@ -2,11 +2,20 @@
 # Copyright: Damien Elmes <anki@ichi2.net>
 # License: GNU AGPL, version 3 or later; http://www.gnu.org/licenses/agpl.html
 
-import time, os, random, stat, datetime, copy
+import pprint
+import re
+import time
+import os
+import random
+import stat
+import datetime
+import copy
+import traceback
+
 from anki.lang import _, ngettext
 from anki.utils import ids2str, fieldChecksum, stripHTML, \
     intTime, splitFields, joinFields, maxID, json
-from anki.hooks import  runFilter
+from anki.hooks import  runFilter, runHook
 from anki.sched import Scheduler
 from anki.models import ModelManager
 from anki.media import MediaManager
@@ -15,9 +24,12 @@ from anki.tags import TagManager
 from anki.consts import *
 from anki.errors import AnkiError
 from anki.sound import stripSounds
-
 import anki.latex # sets up hook
-import anki.cards, anki.notes, anki.template, anki.find
+import anki.cards
+import anki.notes
+import anki.template
+import anki.find
+
 
 defaultConf = {
     # review options
@@ -39,9 +51,12 @@ defaultConf = {
 # this is initialized by storage.Collection
 class _Collection(object):
 
-    def __init__(self, db, server=False):
+    def __init__(self, db, server=False, log=False):
+        self._debugLog = log
         self.db = db
         self.path = db._path
+        self._openLog()
+        self.log(self.path, anki.version)
         self.server = server
         self._lastSave = time.time()
         self.clearUndo()
@@ -57,8 +72,10 @@ class _Collection(object):
             d += datetime.timedelta(hours=4)
             self.crt = int(time.mktime(d.timetuple()))
         self.sched = Scheduler(self)
-        # check for improper shutdown
-        self.cleanup()
+        if not self.conf.get("newBury", False):
+            mod = self.db.mod
+            self.sched.unburyCards()
+            self.db.mod = mod
 
     def name(self):
         n = os.path.splitext(os.path.basename(self.path))[0]
@@ -71,7 +88,7 @@ class _Collection(object):
         (self.crt,
          self.mod,
          self.scm,
-         self.dty,
+         self.dty, # no longer used
          self._usn,
          self.ls,
          self.conf,
@@ -131,7 +148,10 @@ crt=?, mod=?, scm=?, dty=?, usn=?, ls=?, conf=?""",
     def close(self, save=True):
         "Disconnect from DB."
         if self.db:
-            self.cleanup()
+            if not self.conf.get("newBury", False):
+                mod = self.db.mod
+                self.sched.unburyCards()
+                self.db.mod = mod
             if save:
                 self.save()
             else:
@@ -141,6 +161,7 @@ crt=?, mod=?, scm=?, dty=?, usn=?, ls=?, conf=?""",
             self.db.close()
             self.db = None
             self.media.close()
+            self._closeLog()
 
     def reopen(self):
         "Reconnect to DB (after changing threads, etc)."
@@ -148,6 +169,7 @@ crt=?, mod=?, scm=?, dty=?, usn=?, ls=?, conf=?""",
         if not self.db:
             self.db = anki.db.DB(self.path)
             self.media.connect()
+            self._openLog()
 
     def rollback(self):
         self.db.rollback()
@@ -160,35 +182,31 @@ crt=?, mod=?, scm=?, dty=?, usn=?, ls=?, conf=?""",
             if check and not runFilter("modSchema", True):
                 raise AnkiError("abortSchemaMod")
         self.scm = intTime(1000)
+        self.setMod()
 
     def schemaChanged(self):
         "True if schema changed since last sync."
         return self.scm > self.ls
 
-    def setDirty(self):
-        "Signal there are temp. suspended cards that need cleaning up on close."
-        self.dty = True
-
-    def cleanup(self):
-        "Unsuspend any temporarily suspended cards."
-        if self.dty:
-            self.sched.unburyCards()
-            self.dty = False
-
     def usn(self):
         return self._usn if self.server else -1
 
     def beforeUpload(self):
         "Called before a full upload."
-        tbls = "notes", "cards", "revlog", "graves"
+        tbls = "notes", "cards", "revlog"
         for t in tbls:
             self.db.execute("update %s set usn=0 where usn=-1" % t)
+        # we can save space by removing the log of deletions
+        self.db.execute("delete from graves")
         self._usn += 1
         self.models.beforeUpload()
         self.tags.beforeUpload()
         self.decks.beforeUpload()
         self.modSchema()
         self.ls = self.scm
+        # ensure db is compacted before upload
+        self.db.execute("vacuum")
+        self.db.execute("analyze")
         self.close()
 
     # Object creation helpers
@@ -258,6 +276,7 @@ crt=?, mod=?, scm=?, dty=?, usn=?, ls=?, conf=?""",
         strids = ids2str(ids)
         # we need to log these independently of cards, as one side may have
         # more card templates
+        runHook("remNotes", self, ids)
         self._logRem(ids, REM_NOTE)
         self.db.execute("delete from notes where id in %s" % strids)
 
@@ -608,6 +627,10 @@ where c.nid == f.id
             "select id from revlog where cid = ? "
             "order by id desc limit 1", c.id)
         self.db.execute("delete from revlog where id = ?", last)
+        # restore any siblings
+        self.db.execute(
+            "update cards set queue=type,mod=?,usn=? where queue=-2 and nid=?",
+            intTime(), self.usn(), c.nid)
         # and finally, update daily counts
         n = 1 if c.queue == 3 else c.queue
         type = ("new", "lrn", "rev")[n]
@@ -631,6 +654,30 @@ where c.nid == f.id
     # DB maintenance
     ##########################################################################
 
+    def basicCheck(self):
+        "Basic integrity check for syncing. True if ok."
+        # cards without notes
+        if self.db.scalar("""
+select 1 from cards where nid not in (select id from notes) limit 1"""):
+            return
+        # notes without cards or models
+        if self.db.scalar("""
+select 1 from notes where id not in (select distinct nid from cards)
+or mid not in %s limit 1""" % ids2str(self.models.ids())):
+            return
+        # invalid ords
+        for m in self.models.all():
+            # ignore clozes
+            if m['type'] != MODEL_STD:
+                continue
+            if self.db.scalar("""
+select 1 from cards where ord not in %s and nid in (
+select id from notes where mid = ?) limit 1""" %
+                               ids2str([t['ord'] for t in m['tmpls']]),
+                               m['id']):
+                return
+        return True
+
     def fixIntegrity(self):
         "Fix possible problems and rebuild caches."
         problems = []
@@ -647,22 +694,37 @@ select id from notes where mid not in """ + ids2str(self.models.ids()))
                          "Deleted %d notes with missing note type.", len(ids))
                          % len(ids))
             self.remNotes(ids)
-        # cards with invalid ordinal
+        # for each model
         for m in self.models.all():
-            # ignore clozes
-            if m['type'] != MODEL_STD:
-                continue
-            ids = self.db.list("""
+            if m['type'] == MODEL_STD:
+                # model with missing req specification
+                if 'req' not in m:
+                    self.models._updateRequired(m)
+                    problems.append(_("Fixed note type: %s") % m['name'])
+                # cards with invalid ordinal
+                ids = self.db.list("""
 select id from cards where ord not in %s and nid in (
 select id from notes where mid = ?)""" %
-                               ids2str([t['ord'] for t in m['tmpls']]),
-                               m['id'])
+                                   ids2str([t['ord'] for t in m['tmpls']]),
+                                   m['id'])
+                if ids:
+                    problems.append(
+                        ngettext("Deleted %d card with missing template.",
+                                 "Deleted %d cards with missing template.",
+                                 len(ids)) % len(ids))
+                    self.remCards(ids)
+            # notes with invalid field count
+            ids = []
+            for id, flds in self.db.execute(
+                    "select id, flds from notes where mid = ?", m['id']):
+                if (flds.count("\x1f") + 1) != len(m['flds']):
+                    ids.append(id)
             if ids:
                 problems.append(
-                    ngettext("Deleted %d card with missing template.",
-                             "Deleted %d cards with missing template.",
+                    ngettext("Deleted %d note with wrong field count.",
+                             "Deleted %d notes with wrong field count.",
                              len(ids)) % len(ids))
-                self.remCards(ids)
+                self.remNotes(ids)
         # delete any notes with missing cards
         ids = self.db.list("""
 select id from notes where id not in (select distinct nid from cards)""")
@@ -681,6 +743,16 @@ select id from cards where nid not in (select id from notes)""")
                 ngettext("Deleted %d card with missing note.",
                          "Deleted %d cards with missing note.", cnt) % cnt)
             self.remCards(ids)
+        # cards with odue set when it shouldn't be
+        ids = self.db.list("""
+select id from cards where odue > 0 and (type=1 or queue=2) and not odid""")
+        if ids:
+            cnt = len(ids)
+            problems.append(
+                ngettext("Fixed %d card with invalid properties.",
+                         "Fixed %d cards with invalid properties.", cnt) % cnt)
+            self.db.execute("update cards set odue=0 where id in "+
+                ids2str(ids))
         # tags
         self.tags.registerNotes()
         # field cache
@@ -717,3 +789,35 @@ and queue = 0""", intTime(), self.usn())
         self.db.execute("vacuum")
         self.db.execute("analyze")
         self.lock()
+
+    # Logging
+    ##########################################################################
+
+    def log(self, *args, **kwargs):
+        if not self._debugLog:
+            return
+        def customRepr(x):
+            if isinstance(x, basestring):
+                return x
+            return pprint.pformat(x)
+        path, num, fn, y = traceback.extract_stack(
+            limit=2+kwargs.get("stack", 0))[0]
+        buf = u"[%s] %s:%s(): %s" % (intTime(), os.path.basename(path), fn,
+                                     ", ".join([customRepr(x) for x in args]))
+        self._logHnd.write(buf.encode("utf8") + "\n")
+        if os.environ.get("ANKIDEV"):
+            print buf
+
+    def _openLog(self):
+        if not self._debugLog:
+            return
+        lpath = re.sub("\.anki2$", ".log", self.path)
+        if os.path.exists(lpath) and os.path.getsize(lpath) > 10*1024*1024:
+            lpath2 = lpath + ".old"
+            if os.path.exists(lpath2):
+                os.unlink(lpath2)
+            os.rename(lpath, lpath2)
+        self._logHnd = open(lpath, "ab")
+
+    def _closeLog(self):
+        self._logHnd = None
