@@ -3,10 +3,13 @@
 # License: GNU AGPL, version 3 or later; http://www.gnu.org/licenses/agpl.html
 
 import re
-from anki.utils import ids2str, splitFields, joinFields, intTime
-from anki.consts import *
 import sre_constants
 
+from anki.utils import ids2str, splitFields, joinFields, intTime, fieldChecksum, stripHTMLMedia
+from anki.consts import *
+from anki.hooks import *
+
+
 # Find
 ##########################################################################
 
@@ -14,6 +17,21 @@ class Finder(object):
 
     def __init__(self, col):
         self.col = col
+        self.search = dict(
+            added=self._findAdded,
+            card=self._findTemplate,
+            deck=self._findDeck,
+            mid=self._findMid,
+            nid=self._findNids,
+            cid=self._findCids,
+            note=self._findModel,
+            prop=self._findProp,
+            rated=self._findRated,
+            tag=self._findTag,
+            dupe=self._findDupes,
+        )
+        self.search['is'] = self._findCardState
+        runHook("search", self.search)
 
     def findCards(self, query, order=False):
         "Return a list of card ids for QUERY."
@@ -154,26 +172,8 @@ select distinct(n.id) from cards c, notes n where c.nid=n.id and """+preds
             elif ":" in token:
                 cmd, val = token.split(":", 1)
                 cmd = cmd.lower()
-                if cmd == "tag":
-                    add(self._findTag(val, args))
-                elif cmd == "is":
-                    add(self._findCardState(val))
-                elif cmd == "nid":
-                    add(self._findNids(val))
-                elif cmd == "card":
-                    add(self._findTemplate(val))
-                elif cmd == "note":
-                    add(self._findModel(val))
-                elif cmd == "mid":
-                    add(self._findMid(val))
-                elif cmd == "deck":
-                    add(self._findDeck(val))
-                elif cmd == "prop":
-                    add(self._findProp(val))
-                elif cmd == "rated":
-                    add(self._findRated(val))
-                elif cmd == "added":
-                    add(self._findAdded(val))
+                if cmd in self.search:
+                    add(self.search[cmd]((val, args)))
                 else:
                     add(self._findField(cmd, val))
             # normal text search
@@ -239,7 +239,7 @@ select distinct(n.id) from cards c, notes n where c.nid=n.id and """+preds
     # Commands
     ######################################################################
 
-    def _findTag(self, val, args):
+    def _findTag(self, (val, args)):
         if val == "none":
             return 'n.tags = ""'
         val = val.replace("*", "%")
@@ -250,7 +250,7 @@ select distinct(n.id) from cards c, notes n where c.nid=n.id and """+preds
         args.append(val)
         return "n.tags like ?"
 
-    def _findCardState(self, val):
+    def _findCardState(self, (val, args)):
         if val in ("review", "new", "learn"):
             if val == "review":
                 n = 2
@@ -261,13 +261,15 @@ select distinct(n.id) from cards c, notes n where c.nid=n.id and """+preds
             return "type = %d" % n
         elif val == "suspended":
             return "c.queue = -1"
+        elif val == "buried":
+            return "c.queue = -2"
         elif val == "due":
             return """
 (c.queue in (2,3) and c.due <= %d) or
 (c.queue = 1 and c.due <= %d)""" % (
     self.col.sched.today, self.col.sched.dayCutoff)
 
-    def _findRated(self, val):
+    def _findRated(self, (val, args)):
         # days(:optional_ease)
         r = val.split(":")
         try:
@@ -285,7 +287,7 @@ select distinct(n.id) from cards c, notes n where c.nid=n.id and """+preds
         return ("c.id in (select cid from revlog where id>%d %s)" %
                 (cutoff, ease))
 
-    def _findAdded(self, val):
+    def _findAdded(self, (val, args)):
         try:
             days = int(val)
         except ValueError:
@@ -293,7 +295,7 @@ select distinct(n.id) from cards c, notes n where c.nid=n.id and """+preds
         cutoff = (self.col.sched.dayCutoff - 86400*days)*1000
         return "c.id > %d" % cutoff
 
-    def _findProp(self, val):
+    def _findProp(self, (val, args)):
         # extract
         m = re.match("(^.+?)(<=|>=|!=|=|<|>)(.+?$)", val)
         if not m:
@@ -329,17 +331,22 @@ select distinct(n.id) from cards c, notes n where c.nid=n.id and """+preds
         args.append("%"+val+"%")
         return "(n.sfld like ? escape '\\' or n.flds like ? escape '\\')"
 
-    def _findNids(self, val):
+    def _findNids(self, (val, args)):
         if re.search("[^0-9,]", val):
             return
         return "n.id in (%s)" % val
 
-    def _findMid(self, val):
+    def _findCids(self, (val, args)):
+        if re.search("[^0-9,]", val):
+            return
+        return "c.id in (%s)" % val
+
+    def _findMid(self, (val, args)):
         if re.search("[^0-9]", val):
             return
         return "n.mid = %s" % val
 
-    def _findModel(self, val):
+    def _findModel(self, (val, args)):
         ids = []
         val = val.lower()
         for m in self.col.models.all():
@@ -347,7 +354,7 @@ select distinct(n.id) from cards c, notes n where c.nid=n.id and """+preds
                 ids.append(m['id'])
         return "n.mid in %s" % ids2str(ids)
 
-    def _findDeck(self, val):
+    def _findDeck(self, (val, args)):
         # if searching for all decks, skip
         if val == "*":
             return "skip"
@@ -368,7 +375,9 @@ select distinct(n.id) from cards c, notes n where c.nid=n.id and """+preds
         else:
             # wildcard
             ids = set()
+            # should use re.escape in the future
             val = val.replace("*", ".*")
+            val = val.replace("+", "\\+")
             for d in self.col.decks.all():
                 if re.match("(?i)"+val, d['name']):
                     ids.update(dids(d['id']))
@@ -377,7 +386,7 @@ select distinct(n.id) from cards c, notes n where c.nid=n.id and """+preds
         sids = ids2str(ids)
         return "c.did in %s or c.odid in %s" % (sids, sids)
 
-    def _findTemplate(self, val):
+    def _findTemplate(self, (val, args)):
         # were we given an ordinal number?
         try:
             num = int(val) - 1
@@ -429,7 +438,22 @@ where mid in %s and flds like ? escape '\\'""" % (
             except sre_constants.error:
                 return
         if not nids:
+            return "0"
+        return "n.id in %s" % ids2str(nids)
+
+    def _findDupes(self, (val, args)):
+        # caller must call stripHTMLMedia on passed val
+        try:
+            mid, val = val.split(",", 1)
+        except OSError:
             return
+        csum = fieldChecksum(val)
+        nids = []
+        for nid, flds in self.col.db.execute(
+                "select id, flds from notes where mid=? and csum=?",
+                mid, csum):
+            if stripHTMLMedia(splitFields(flds)[0]) == val:
+                nids.append(nid)
         return "n.id in %s" % ids2str(nids)
 
 # Find and replace
@@ -498,7 +522,7 @@ def fieldNames(col, downcase=True):
 
 # Find duplicates
 ##########################################################################
-
+# returns array of ("dupestr", [nids])
 def findDupes(col, fieldName, search=""):
     # limit search to notes with applicable field name
     if search:
@@ -511,13 +535,20 @@ def findDupes(col, fieldName, search=""):
     def ordForMid(mid):
         if mid not in fields:
             model = col.models.get(mid)
-            fields[mid] = col.models.fieldMap(model)[fieldName][0]
+            for c, f in enumerate(model['flds']):
+                if f['name'].lower() == fieldName.lower():
+                    fields[mid] = c
+                    break
         return fields[mid]
     for nid, mid, flds in col.db.all(
         "select id, mid, flds from notes where id in "+ids2str(
             col.findNotes(search))):
         flds = splitFields(flds)
-        val = flds[ordForMid(mid)]
+        ord = ordForMid(mid)
+        if ord is None:
+            continue
+        val = flds[ord]
+        val = stripHTMLMedia(val)
         # empty does not count as duplicate
         if not val:
             continue
