@@ -2,11 +2,16 @@
 # Copyright: Damien Elmes <anki@ichi2.net>
 # License: GNU AGPL, version 3 or later; http://www.gnu.org/licenses/agpl.html
 
-import time, datetime, json
+from __future__ import division
+import time
+import datetime
+import json
+
 import anki.js
 from anki.utils import fmtTimeSpan, ids2str
 from anki.lang import _, ngettext
 
+
 # Card stats
 ##########################################################################
 
@@ -55,6 +60,8 @@ class CardStats(object):
         self.addLine(_("Card Type"), c.template()['name'])
         self.addLine(_("Note Type"), c.model()['name'])
         self.addLine(_("Deck"), self.col.decks.name(c.did))
+        self.addLine(_("Note ID"), c.nid)
+        self.addLine(_("Card ID"), c.id)
         self.txt += "</table>"
         return self.txt
 
@@ -165,6 +172,16 @@ from revlog where id > ? """+lim, (self.col.sched.dayCutoff-86400)*1000)
         b += "<br>"
         b += (_("Learn: %(a)s, Review: %(b)s, Relearn: %(c)s, Filtered: %(d)s")
               % dict(a=bold(lrn), b=bold(rev), c=bold(relrn), d=bold(filt)))
+        # mature today
+        mcnt, msum = self.col.db.first("""
+select count(), sum(case when ease = 1 then 0 else 1 end) from revlog
+where lastIvl >= 21 and id > ?"""+lim, (self.col.sched.dayCutoff-86400)*1000)
+        b += "<br>"
+        if mcnt:
+            b += _("Correct answers on mature cards: %(a)d/%(b)d (%(c).1f%%)") % dict(
+                a=msum, b=mcnt, c=(msum / float(mcnt) * 100))
+        else:
+            b += _("No mature cards were studied today.")
         return b
 
     # Due and cumulative due
@@ -203,8 +220,8 @@ from revlog where id > ? """+lim, (self.col.sched.dayCutoff-86400)*1000)
             xaxis['max'] = end-0.5
         txt += self._graph(id="due", data=data,
                            ylabel2=_("Cumulative Cards"), conf=dict(
-                xaxis=xaxis, yaxes=[dict(), dict(
-                    tickDecimals=0, position="right")]))
+                xaxis=xaxis, yaxes=[dict(min=0), dict(
+                    min=0, tickDecimals=0, position="right")]))
         txt += self._dueInfo(tot, len(totd)*chunk)
         return txt
 
@@ -258,7 +275,7 @@ group by day order by day""" % (self._limit(), lim),
         d = data
         conf = dict(
             xaxis=dict(tickDecimals=0, max=0.5),
-            yaxes=[dict(), dict(position="right")])
+            yaxes=[dict(min=0), dict(position="right",min=0)])
         if days is not None:
             conf['xaxis']['min'] = -days+0.5
         def plot(id, data, ylabel, ylabel2):
@@ -313,7 +330,7 @@ group by day order by day""" % (self._limit(), lim),
                 period = 1
             else:
                 period = max(
-                    1, (self.col.sched.dayCutoff - (t/1000)) / 86400)
+                    1, int(1+((self.col.sched.dayCutoff - (t/1000)) / 86400)))
         i = []
         self._line(i, _("Days studied"),
                    _("<b>%(pct)d%%</b> (%(x)s of %(y)s)") % dict(
@@ -330,14 +347,16 @@ group by day order by day""" % (self._limit(), lim),
             tot *= 60
         self._line(i, _("Average for days studied"), self._avgDay(
             tot, studied, unit))
-        self._line(i, _("If you studied every day"), self._avgDay(
-            tot, period, unit))
+        if studied != period:
+            # don't display if you did study every day
+            self._line(i, _("If you studied every day"), self._avgDay(
+                tot, period, unit))
         if total and tot:
             perMin = total / float(tot)
             perMin = ngettext("%d card/minute", "%d cards/minute", perMin) % perMin
             self._line(
                 i, _("Average answer time"),
-                "%0.1fs (%s)" % ((tot*60)/total, perMin))
+                _("%(a)0.1fs (%(b)s)") % dict(a=(tot*60)/total, b=perMin))
         return self._lineTbl(i), int(tot)
 
     def _splitRepData(self, data, spec):
@@ -768,7 +787,7 @@ font-weight: bold;
 
 <td>
 <center><div id=%(id)sLegend></div></center>
-<div id="%(id)s" style="width:%(w)s; height:%(h)s;"></div>
+<div id="%(id)s" style="width:%(w)spx; height:%(h)spx;"></div>
 </td>
 
 <td><div style="width: 150px; text-align: center; position:absolute;
