@@ -1,14 +1,20 @@
 # Copyright: Damien Elmes <anki@ichi2.net>
 # License: GNU AGPL, version 3 or later; http://www.gnu.org/licenses/agpl.html
 
+from __future__ import division
+import socket
+import time
+import traceback
+import gc
+
 from aqt.qt import *
-import   socket, time, traceback, gc
 import aqt
 from anki import Collection
 from anki.sync import Syncer, RemoteServer, FullSyncer, MediaSyncer, \
     RemoteMediaServer
 from anki.hooks import addHook, remHook
-from aqt.utils import tooltip, askUserDialog, showWarning, showText
+from aqt.utils import tooltip, askUserDialog, showWarning, showText, showInfo
+
 
 # Sync manager
 ######################################################################
@@ -25,6 +31,7 @@ class SyncManager(QObject):
             auth = self._getUserPass()
             if not auth:
                 return
+            self.pm.profile['syncUser'] = auth[0]
             self._sync(auth)
         else:
             self._sync()
@@ -32,6 +39,8 @@ class SyncManager(QObject):
     def _sync(self, auth=None):
         # to avoid gui widgets being garbage collected in the worker thread,
         # run gc in advance
+        self._didFullUp = False
+        self._didError = False
         gc.collect()
         # create the thread, setup signals and start running
         t = self.thread = SyncThread(
@@ -47,13 +56,27 @@ class SyncManager(QObject):
             self.mw.app.processEvents()
             self.thread.wait(100)
         self.mw.progress.finish()
+        if self.thread.syncMsg:
+            showText(self.thread.syncMsg)
+        if self.thread.uname:
+            self.pm.profile['syncUser'] = self.thread.uname
+        def delayedInfo():
+            if self._didFullUp and not self._didError:
+                showInfo(_("""\
+Your collection was successfully uploaded to AnkiWeb.
+
+If you use any other devices, please sync them now, and choose \
+to download the collection you have just uploaded from this computer. \
+After doing so, future reviews and added cards will be merged \
+automatically."""))
+        self.mw.progress.timer(1000, delayedInfo, False)
 
     def _updateLabel(self):
         self.mw.progress.update(label="%s\n%s" % (
             self.label,
             _("%(a)dkB up, %(b)dkB down") % dict(
-                a=self.sentBytes/1024,
-                b=self.recvBytes/1024)))
+                a=self.sentBytes // 1024,
+                b=self.recvBytes // 1024)))
 
     def onEvent(self, evt, *args):
         pu = self.mw.progress.update
@@ -71,11 +94,15 @@ class SyncManager(QObject):
             self.pm.save()
         elif evt == "offline":
             tooltip(_("Syncing failed; internet offline."))
+        elif evt == "upbad":
+            self._didFullUp = False
+            self._checkFailed()
         elif evt == "sync":
             m = None; t = args[0]
             if t == "login":
                 m = _("Syncing...")
             elif t == "upload":
+                self._didFullUp = True
                 m = _("Uploading to AnkiWeb...")
             elif t == "download":
                 m = _("Downloading from AnkiWeb...")
@@ -90,10 +117,17 @@ Please visit AnkiWeb, upgrade your deck, then try again."""))
                 self.label = m
                 self._updateLabel()
         elif evt == "error":
+            self._didError = True
             showText(_("Syncing failed:\n%s")%
                      self._rewriteError(args[0]))
         elif evt == "clockOff":
             self._clockOff()
+        elif evt == "checkFailed":
+            self._checkFailed()
+        elif evt == "mediaSanity":
+            showWarning(_("""\
+A problem occurred while syncing media. Please use Tools>Check Media, then \
+sync again to correct the issue."""))
         elif evt == "noChanges":
             pass
         elif evt == "fullSync":
@@ -115,29 +149,35 @@ and try again.""")
             return _("""\
 The connection to AnkiWeb timed out. Please check your network \
 connection and try again.""")
-        elif "500" in err:
+        elif "code: 500" in err:
             return _("""\
 AnkiWeb encountered an error. Please try again in a few minutes, and if \
 the problem persists, please file a bug report.""")
-        elif "501" in err:
+        elif "code: 501" in err:
             return _("""\
 Please upgrade to the latest version of Anki.""")
         # 502 is technically due to the server restarting, but we reuse the
         # error message
-        elif "502" in err or "503" in err or "504" in err:
+        elif "code: 502" in err:
+            return _("AnkiWeb is under maintenance. Please try again in a few minutes.")
+        elif "code: 503" in err:
             return _("""\
 AnkiWeb is too busy at the moment. Please try again in a few minutes.""")
-        elif "409" in err:
-            return _("A previous sync failed; please try again in a few minutes.")
-        elif "10061" in err or "10013" in err:
+        elif "code: 504" in err:
+            return _("504 gateway timeout error received. Please try temporarily disabling your antivirus.")
+        elif "code: 409" in err:
+            return _("Only one client can access AnkiWeb at a time. If a previous sync failed, please try again in a few minutes.")
+        elif "10061" in err or "10013" in err or "10053" in err:
             return _(
                 "Antivirus or firewall software is preventing Anki from connecting to the internet.")
-        elif "407" in err:
+        elif "Unable to find the server" in err:
+            return _(
+                "Server not found. Either your connection is down, or antivirus/firewall "
+                "software is blocking Anki from connecting to the internet.")
+        elif "code: 407" in err:
             return _("Proxy authentication required.")
-        elif "collection sanity check failed" in err:
-            return _("After syncing, the collection was in an inconsistent \
-state. To fix this problem, Anki will force a full sync. Please sync again, and \
-choose which side you would like to keep.")
+        elif "code: 413" in err:
+            return _("Your collection or a media file is too large to sync.")
         return err
 
     def _getUserPass(self):
@@ -186,8 +226,16 @@ Your decks here and on AnkiWeb differ in such a way that they can't \
 be merged together, so it's necessary to overwrite the decks on one \
 side with the decks from the other.
 
-Do you want to upload the decks from here, or download the decks \
-from AnkiWeb?"""),
+If you choose download, Anki will download the collection from AnkiWeb, \
+and any changes you have made on your computer since the last sync will \
+be lost.
+
+If you choose upload, Anki will upload your collection to AnkiWeb, and \
+any changes you have made on AnkiWeb or your other devices since the \
+last sync to this device will be lost.
+
+After all devices are in sync, future reviews and added cards can be merged \
+automatically."""),
                 [_("Upload to AnkiWeb"),
                  _("Download from AnkiWeb"),
                  _("Cancel")])
@@ -205,6 +253,11 @@ from AnkiWeb?"""),
 Syncing requires the clock on your computer to be set correctly. Please \
 fix the clock and try again."""))
 
+    def _checkFailed(self):
+        showWarning(_("""\
+Your collection is in an inconsistent state. Please run Tools>\
+Check Database, then sync again."""))
+
     def badUserPass(self):
         aqt.preferences.Preferences(self, self.pm.profile).dialog.tabWidget.\
                                          setCurrentIndex(1)
@@ -222,8 +275,12 @@ class SyncThread(QThread):
         self.media = media
 
     def run(self):
+        # init this first so an early crash doesn't cause an error
+        # in the main thread
+        self.syncMsg = ""
+        self.uname = ""
         try:
-            self.col = Collection(self.path)
+            self.col = Collection(self.path, log=True)
         except:
             self.fireEvent("corrupt")
             return
@@ -280,12 +337,9 @@ class SyncThread(QThread):
             ret = self.client.sync()
         except Exception, e:
             log = traceback.format_exc()
-            try:
-                err = unicode(e[0], "utf8", "ignore")
-            except:
-                # number, exception with no args, etc
-                err = ""
-            if "Unable to find the server" in err:
+            err = repr(str(e))
+            if ("Unable to find the server" in err or
+                "Errno 2" in err):
                 self.fireEvent("offline")
             else:
                 if not err:
@@ -298,6 +352,8 @@ class SyncThread(QThread):
             return self.fireEvent("badAuth")
         elif ret == "clockOff":
             return self.fireEvent("clockOff")
+        elif ret == "basicCheckFailed" or ret == "sanityCheckFailed":
+            return self.fireEvent("checkFailed")
         # note mediaUSN for later
         self.mediaUsn = self.client.mediaUsn
         # full sync?
@@ -306,8 +362,14 @@ class SyncThread(QThread):
         # save and note success state
         if ret == "noChanges":
             self.fireEvent("noChanges")
-        else:
+        elif ret == "success":
             self.fireEvent("success")
+        elif ret == "serverAbort":
+            self.fireEvent("error", self.client.syncMsg)
+        else:
+            self.fireEvent("error", "Unknown sync return code.")
+        self.syncMsg = self.client.syncMsg
+        self.uname = self.client.uname
         # then move on to media sync
         self._syncMedia()
 
@@ -327,7 +389,8 @@ class SyncThread(QThread):
             return
         self.client = FullSyncer(self.col, self.hkey, self.server.con)
         if f == "upload":
-            self.client.upload()
+            if not self.client.upload():
+                self.fireEvent("upbad")
         else:
             self.client.download()
         # reopen db and move on to media sync
@@ -342,6 +405,8 @@ class SyncThread(QThread):
         ret = self.client.sync(self.mediaUsn)
         if ret == "noChanges":
             self.fireEvent("noMediaChanges")
+        elif ret == "sanityCheckFailed":
+            self.fireEvent("mediaSanity")
         else:
             self.fireEvent("mediaSuccess")
 
@@ -353,7 +418,7 @@ class SyncThread(QThread):
 ######################################################################
 
 CHUNK_SIZE = 65536
-import httplib, httplib2, errno
+import httplib, httplib2
 from cStringIO import StringIO
 from anki.hooks import runHook
 
@@ -367,6 +432,8 @@ def _incrementalSend(self, data):
             raise httplib.NotConnected()
     # if it's not a file object, make it one
     if not hasattr(data, 'read'):
+        if isinstance(data, unicode):
+            data = data.encode("utf8")
         data = StringIO(data)
     while 1:
         block = data.read(CHUNK_SIZE)
@@ -378,6 +445,10 @@ def _incrementalSend(self, data):
 httplib.HTTPConnection.send = _incrementalSend
 
 # receiving in httplib2
+# this is an augmented version of httplib's request routine that:
+# - doesn't assume requests will be tried more than once
+# - calls a hook for each chunk of data so we can update the gui
+# - retries only when keep-alive connection is closed
 def _conn_request(self, conn, request_uri, method, body, headers):
     for i in range(2):
         try:
@@ -394,38 +465,20 @@ def _conn_request(self, conn, request_uri, method, body, headers):
             conn.close()
             raise
         except socket.error, e:
-            err = 0
-            if hasattr(e, 'args'):
-                err = getattr(e, 'args')[0]
-            else:
-                err = e.errno
-            if err == errno.ECONNREFUSED: # Connection refused
-                raise
+            conn.close()
+            raise
         except httplib.HTTPException:
-            # Just because the server closed the connection doesn't apparently mean
-            # that the server didn't send a response.
-            if conn.sock is None:
-                if i == 0:
-                    conn.close()
-                    conn.connect()
-                    continue
-                else:
-                    conn.close()
-                    raise
-            if i == 0:
-                conn.close()
-                conn.connect()
-                continue
-            pass
+            conn.close()
+            raise
         try:
             response = conn.getresponse()
+        except httplib.BadStatusLine:
+            print "retry bad line"
+            conn.close()
+            conn.connect()
+            continue
         except (socket.error, httplib.HTTPException):
-            if i == 0:
-                conn.close()
-                conn.connect()
-                continue
-            else:
-                raise
+            raise
         else:
             content = ""
             if method == "HEAD":
@@ -442,7 +495,6 @@ def _conn_request(self, conn, request_uri, method, body, headers):
             response = httplib2.Response(response)
             if method != "HEAD":
                 content = httplib2._decompressContent(response, content)
-        break
-    return (response, content)
+        return (response, content)
 
 httplib2.Http._conn_request = _conn_request
